/*
** Module   : registry.c
** Abstract : implementation for Windows registry access functions.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 EVL 20130618 Created initial version to implement Windows specifics for registry related
**                  environment sccess operations.
*/

#include <windows.h>

#include "com_goldencode_p2j_util_FileSystemDaemon.h"

static const char * NAME_SEP = "\\\\";

/**
* Utility method that will return the value of the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to get.
* 
* @return A string representing the registry key value if exists or NULL pointer if the key does
*         not exist in the Windows registry.
*/
JNIEXPORT jstring JNICALL
   Java_com_goldencode_p2j_util_Registry_getKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key)
{
   char * result = NULL;

   // Construct initial parameters for Windows API call
   const char *base_name = (*env)->GetStringUTFChars(env, base, NULL);
   const char *sect_name = (*env)->GetStringUTFChars(env, section, NULL);
   const char *key_name = (*env)->GetStringUTFChars(env, key, NULL);
   
   // Reconstruct the handle to the registry hive
   HKEY hKey = HKEY_CURRENT_USER;
   // TODO: Implement handling the other cases of base_name variable
   
   int keyPathLength = 0;
   // Check the memory size to hold the full key name
   // TODO: Implement handling different missing parts either section or key name combinations
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      keyPathLength += strlen(sect_name) + strlen(NAME_SEP);
   }
   if (key_name != NULL && strlen(key_name) > 0)
   {
      keyPathLength += strlen(key_name);
   }
   
   // Allocate memory to registry key name
   char * fullKeyName = malloc(sizeof(char) * keyPathLength);
   // And initialize with empty
   strcpy(fullKeyName, "");
   // Form the final key name
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      strcat(fullKeyName, sect_name);
      strcat(fullKeyName, NAME_SEP);
   }   
   if (key_name != NULL && strlen(key_name) > 0)
   {
      strcat(fullKeyName, key_name);
   }
   
   DWORD rc;
   DWORD cbData = 0;
   
   // Find out how much memory we need to store the result
   rc = RegQueryValueEx(hKey, fullKeyName, NULL, NULL, NULL, &cbData);
   if (rc == ERROR_SUCCESS && cbData > 0)
   {
      result = malloc(sizeof(char) * cbData);
      if (result != NULL)
      {
         rc = RegQueryValueEx(hKey, fullKeyName, NULL, NULL, (LPBYTE)result, &cbData);
      }
   }

   // Release local memory
   free(fullKeyName);
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, key, key_name);
   (*env)->ReleaseStringUTFChars(env, section, sect_name);
   (*env)->ReleaseStringUTFChars(env, base, base_name);
   
   if (result != NULL)
   {
      // We need this additional variable to be able to free local memory allocated for
      // Windows API call and to prevent memory leak
      jstring res = (*env)->NewStringUTF(env, result);
      free(result);
      return res;
   }
   else
   {
      return NULL;
   }   
}

/**
* Utility method that will set the new value for the registry key for given section name. 
* 
* @param  env 
*         pointer to a structure that provides access to JNI environment
* @param  cls 
*         reference to the Java class who called this function
*         (from static context)
* @param  base 
*         The name of the registry base section to scan(HKEY_* hives level).
* @param  section 
*         The name of the registry section to scan.
* @param  key 
*         The name of the registry key within given section whose value to change.
* @param  value 
*         The new value for the registry entry to set for given section and given key.
* 
* @return TRUE in case of success, FALSE otherwise.
*/
JNIEXPORT jboolean JNICALL
   Java_com_goldencode_p2j_util_Registry_setKeyVal(JNIEnv* env,
                                                   jclass  cls,
                                                   jstring base,
                                                   jstring section,
                                                   jstring key,
                                                   jstring value)
{
   jboolean res = FALSE;
   // Construct initial parameters for Windows API call
   const char *base_name = (*env)->GetStringUTFChars(env, base, NULL);
   const char *sect_name = (*env)->GetStringUTFChars(env, section, NULL);
   const char *key_name = (*env)->GetStringUTFChars(env, key, NULL);
   const char *key_value = (*env)->GetStringUTFChars(env, value, NULL);
   
   // Reconstruct the handle to the registry hive
   HKEY hKey = HKEY_CURRENT_USER;
   // TODO: Implement handling the other cases of base_name variable
   
   int keyPathLength = 0;
   // Check the memory size to hold the full key name
   // TODO: Implement handling different missing parts either section or key name combinations
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      keyPathLength += strlen(sect_name) + strlen(NAME_SEP);
   }
   if (key_name != NULL && strlen(key_name) > 0)
   {
      keyPathLength += strlen(key_name);
   }
   
   // Allocate memory to registry key name
   char * fullKeyName = malloc(sizeof(char) * keyPathLength);
   // And initialize with empty
   strcpy(fullKeyName, "");
   // Form the final key name
   if (sect_name != NULL && strlen(sect_name) > 0)
   {
      strcat(fullKeyName, sect_name);
      strcat(fullKeyName, NAME_SEP);
   }   
   if (key_name != NULL && strlen(key_name) > 0)
   {
      strcat(fullKeyName, key_name);
   }

   DWORD rc;
   DWORD cbData = 0;
   
   if (key_value != NULL && strlen(key_value) > 0)
   {
      // The size of data to be stored should include null terminator for string data
      cbData = strlen(key_value) + 1;
      rc = RegSetValueEx(hKey, fullKeyName, 0, REG_SZ, (LPBYTE)key_value, cbData);
      if (rc == ERROR_SUCCESS)
      {
         res = TRUE;
      }
   }
   
   // Release local memory
   free(fullKeyName);
   // Release Java memory
   (*env)->ReleaseStringUTFChars(env, value, key_value);
   (*env)->ReleaseStringUTFChars(env, key, key_name);
   (*env)->ReleaseStringUTFChars(env, section, sect_name);
   (*env)->ReleaseStringUTFChars(env, base, base_name);
   
   return res;
}

